import os.path
import dask.array as da
import sqlite3
import numpy as np
import numpy.typing as npt
import pickle


class AnnotationRecordManager:
    """
    Manages useful info to be referenced in the annotation process
    Saves edit history for undo/redo support

    Manager states:
    - *annotate_edit table
    - *metadata table (pointers to annotate_edit)
    - *arrays
    - undo_stack (on the top are the latest undo actions)
    * means persisted above
    """

    def __init__(self, save_path: str):
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        db_path = f'{save_path}/annotation_record.db'
        self.conn = sqlite3.connect(db_path)
        self.cursor = self.conn.cursor()

        self.masks = None
        self.mask_name_to_id = {}
        self.undo_stack = []  # empty if user has not done any undo; used for re-doing actions

        self.cursor.execute(f'''
        CREATE TABLE IF NOT EXISTS metadata (
            id TEXT PRIMARY KEY,
            count INTEGER NOT NULL
        );
        ''')

        self.cursor.execute(f'''
        INSERT INTO metadata (id, count) VALUES ("num_sessions", 0) 
        ON CONFLICT(id) DO NOTHING;
        ''')

        self.cursor.execute(f'''
        CREATE TABLE IF NOT EXISTS session (
            sid INTEGER PRIMARY KEY,
            stime DATE NOT NULL,
            etime DATE
        );
        ''')

        self.cursor.execute(f'''
        CREATE TABLE IF NOT EXISTS annotation_edit (
            mid INTEGER,
            torder INTEGER,
            sid INTEGER REFERENCES session(sid) NOT NULL,
            position TEXT NOT NULL,
            size TEST NOT NULL,
            content_before TEXT NOT NULL,
            content_after TEXT NOT NULL,
            PRIMARY KEY (mid, torder)
        )
        ''')

        print(self.get_attr_count('num_sessions'))
        self.sid = self.inc_attr_count('num_sessions')
        print(self.get_attr_count('num_sessions'))

        self.cursor.execute('''
        INSERT INTO session VALUES (?, DATE('NOW'), ?)
        ON CONFLICT(sid) DO NOTHING;
        ''', (self.sid, None))  # when database connection closes, we will update the end time

    def get_sid(self) -> int:
        return self.sid

    def register_masks(self, masks: list[dict]):
        """Initialize the masks and the corresponding database information

        Not called on initialization because masks requires accessing database information to initialize
        In additional to necessary fields, masks dictionaries may have "on_update" and "on_close" callbacks
        defined.

        Args:
            masks: Dictionaries with fields "name", "arr", "path", "is_numpy" defined
        """
        self.masks = masks
        for i in range(len(masks)):
            mask_info = masks[i]
            for key in ("name", "arr", "path", "is_numpy"):
                assert key in mask_info, f'Key {key} missing in masks array entry {i}!'
        self.mask_name_to_id = {masks[i]['name']: i for i in range(len(masks))}
        mask_names = [mask_info['name'] for mask_info in masks]
        mask_init_values = ','.join(f'("{name}", 0)' for name in mask_names)

        self.cursor.execute(f'''
        INSERT INTO metadata (id, count) VALUES
        {mask_init_values}
        ON CONFLICT(id) DO NOTHING;
        ''')

        print('print database states on register:')
        self.print_table('metadata')
        self.print_table('session')

    def swap_region(self,
                    mask_id: npt.NDArray | da.Array,
                    stpt: npt.NDArray[np.int32],
                    edit_mask: npt.NDArray
                    ) -> npt.NDArray:
        mask = self.masks[mask_id]['arr']
        stpt = stpt.tolist()
        slices = tuple(slice(stpt[i], stpt[i] + edit_mask.shape[i]) for i in range(len(stpt)))
        swapped_region = mask[slices]
        mask[slices] = edit_mask
        if 'on_update' in self.masks[mask_id]:
            self.masks[mask_id]['on_update']()

        return swapped_region

    def do_edit_mask(self, mask_name: str, stpt: npt.NDArray[np.int32], edit_mask: npt.NDArray):
        self.clear_undo_stack()
        mask_id = self.mask_name_to_id[mask_name]

        edit_region_size = np.array(edit_mask.shape, dtype=np.int32)
        swapped_region = self.swap_region(mask_id, stpt, edit_mask)
        order = self.inc_attr_count(mask_name, decrease=False)

        self.insert_row(mask_id, order, stpt, edit_region_size, swapped_region, edit_mask)

    def undo_edit_mask(self, mask_name: str):
        mask_id = self.mask_name_to_id[mask_name]
        order = self.get_attr_count(mask_name)
        if order == 0:
            return

        result = self.pop_row(mask_id, order, delete=False)
        _, _, _, position, _, content_before, _ = result
        self.undo_stack.append((mask_id, order))
        self.swap_region(mask_id, position, content_before)

        self.inc_attr_count(mask_name, decrease=True)

    def redo_edit_mask(self):
        if len(self.undo_stack) == 0:
            return
        mask_id, order = self.undo_stack.pop()
        mask_name = self.masks[mask_id]['name']
        result = self.pop_row(mask_id, order, delete=False)
        _, _, _, position, _, _, content_after = result
        self.swap_region(mask_id, position, content_after)

        self.inc_attr_count(mask_name, decrease=False)

    def clear_undo_stack(self):
        while len(self.undo_stack) > 0:
            mask_id, order = self.undo_stack.pop()
            self.pop_row(mask_id, order, delete=True)

    def insert_row(self,
                   mask_id: int,
                   order: int,
                   stpt: npt.NDArray[np.int32],
                   edit_region_size: npt.NDArray[np.int32],
                   content_before: npt.NDArray,
                   content_after: npt.NDArray):
        self.cursor.execute('''
                        INSERT INTO annotation_edit VALUES
                        (?, ?, ?, ?, ?, ?, ?)
                        ''', (
            mask_id,
            order,
            self.sid,
            pickle.dumps(stpt),
            pickle.dumps(edit_region_size),
            pickle.dumps(content_before),
            pickle.dumps(content_after)
        ))

    def pop_row(self, mask_id: int, order: int, delete: bool = True) -> tuple:
        self.cursor.execute('''
        SELECT * FROM annotate_edit
        WHERE mid = ? AND torder = ?;
        ''', (mask_id, order))

        result = self.cursor.fetchone()
        if delete:
            self.cursor.execute('''
            DELETE FROM annotate_edit WHERE mid = ? AND torder = ?
            ''', (mask_id, order))
        for idx in (3, 4, 5, 6):
            result[idx] = pickle.loads(result[idx])  # these are numpy arrays

        return result

    def get_attr_count(self, attr_name: str) -> int:
        self.cursor.execute('''
        SELECT count FROM metadata WHERE id = ?
        ''', (attr_name,))
        result = self.cursor.fetchone()[0]
        assert isinstance(result, int), f'Expected int, got {type(result)}'
        return result

    def inc_attr_count(self, attr_name: str, decrease: bool = False) -> int:
        self.cursor.execute('''
        SELECT count FROM metadata WHERE id = ?
        ''', (attr_name,))
        result = self.cursor.fetchone()[0]
        assert isinstance(result, int), f'Expected int, got {type(result)}'
        if decrease:
            result -= 1
        else:
            result += 1
        self.cursor.execute('''
        UPDATE metadata SET count = ? WHERE id = ?;
        ''', (result, attr_name))
        return result

    def print_table(self, tname: str):
        print('-' * 30)
        print(f'{tname}')
        self.cursor.execute(f"SELECT * FROM {tname}")
        print(self.cursor.fetchall())

    def commit(self):
        self.clear_undo_stack()
        self.conn.commit()

    def close(self):
        self.cursor.execute('''
        UPDATE session
        SET etime = DATE('NOW')
        WHERE sid = ?
        ''', (self.sid,))  # when database connection closes, we will update the end time
        self.clear_undo_stack()  # commit is done here

        for mask_id in range(len(self.masks)):
            if 'on_close' in self.masks[mask_id]:
                self.masks[mask_id]['on_close']()
        self.cursor.close()
        self.conn.close()




from dataclasses import dataclass
from typing import Any, Callable

import napari
from napari.layers import Layer
import numpy as np
import numpy.typing as npt
import dask.array as da
import json
from cvpl_tools.ome_zarr.napari.annotation_record_manager import AnnotationRecordManager
from napari.utils.events import Event
import cvpl_tools.im.algorithms as np_algorithms
import scipy
import uuid


def are_instersect_slices(slicesA: tuple, slicesB: tuple):
    for i in range(len(slicesA)):
        sA, sB = slicesA[i], slicesB[i]
        if sA.start >= sB.stop or sB.start >= sA.stop:
            return False
    return True


def take_intersect(slicesA: tuple, slicesB: tuple):
    new_s = tuple(slice(max(slicesA[k].start, slicesB[k].start), min(slicesA[k].stop, slicesB[k].stop))
                  for k in range(len(slicesA)))
    return new_s


def are_in_slices(pts: npt.NDArray[np.int32], slices: tuple) -> npt.NDArray[bool]:
    in_slices = np.ones(pts.shape[0], dtype=bool)
    for i in range(len(slices)):
        s = slices[i]
        X = pts[:, i]
        in_slices = in_slices & (X >= s.start) & (X < s.stop)
    return in_slices


def merge_intersecting_slices(slices, ids=None) -> tuple[list, list]:
    """Returns the merged slices and their corresponding ids"""
    slices = list(slices)
    if ids is None:
        ids = [[i] for i in range(len(slices))]
    else:
        ids = [[i] for i in ids]
    assert len(slices) == len(ids), 'Each slice tuple corresponds to an id'
    if len(slices) <= 1:
        return slices, ids

    i = 0
    while i < len(slices) - 1:
        si = slices[i]
        intersect_flag = False
        j = i + 1
        while j < len(slices):
            sj = slices[j]
            if are_instersect_slices(si, sj):
                intersect_flag = True
                break
            j += 1
        if intersect_flag:
            sj = slices.pop(j)
            slices.pop(i)
            new_s = tuple(slice(min(si[k].start, sj[k].start), max(si[k].stop, sj[k].stop)) for k in range(len(si)))
            slices.append(new_s)

            idj = ids.pop(j)
            idi = ids.pop(i)
            ids.append(idi + idj)
        else:
            i += 1
    return slices, ids


def construct_slices(pts: npt.NDArray[np.int32] | list[tuple, ...], width: int):
    if isinstance(pts, np.ndarray):
        pts = pts.tolist()
    ndim = len(pts[0])
    slices = [tuple(slice(pt[i] - width, pt[i] + width + 1) for i in range(ndim)) for pt in pts]
    slices, result_pts = merge_intersecting_slices(slices)
    for i in range(len(result_pts)):
        pts_ids = result_pts[i]
        result_pts[i] = np.array([pts[i] for i in pts_ids], dtype=np.int32)
    return slices, result_pts


@dataclass
class LblInfo:
    thres_arr: npt.NDArray[np.uint8]
    lbl_im: npt.NDArray[np.int32]
    nlbl: int
    cnt_slices: list
    next_pt_id: int

    def __init__(self, thres_arr, next_pt_id: int):
        self.thres_arr = thres_arr
        self.lbl_im, self.nlbl = scipy.ndimage.label(thres_arr)
        self.cnt_slices = list(scipy.ndimage.find_objects(self.lbl_im))
        self.next_pt_id = next_pt_id

    def assign_id(self):
        pt_id = self.next_pt_id
        self.next_pt_id += 1
        return pt_id


class UserInputBuffer:
    """Manages the last few drawn contours"""

    def __init__(self,
                 viewer: napari.Viewer,
                 record_manager: AnnotationRecordManager,
                 masks,
                 name='user_input',
                 bind_key: str | None = None):
        self.im_arr = masks[0]['arr']
        self.im_layer = masks[0]['layer']
        self.ndim = self.im_arr.ndim
        uin_arr: npt.NDArray[np.uint8] = np.zeros(self.im_arr.shape, dtype=np.uint8)
        self.uin_layer: Layer = viewer.add_labels(uin_arr, name=name, scale=self.im_layer.scale)
        self.uin_points_layer: Layer = viewer.add_points(ndim=3, name=f'{name}_shape', size=1,
                                                         scale=self.im_layer.scale)
        self.uin_points_layer.mode = 'add'  # default to add points
        self.record_manager = record_manager
        self.width = 40

        self.last_pts: npt.NDArray[np.int32] = None
        self.lblinfo: LblInfo = None
        self.pt_to_paintid = None
        self.pt_to_cnt_idx = None
        self.cnt_idx_to_pt = None
        self.cnt_idx_to_pt_history = None

        self.on_points_changed = self.uin_points_layer.events.data.connect(self.on_points_changed)
        self.on_push = None

        if bind_key is not None:
            @viewer.bind_key(bind_key)
            def when_push(e):
                if self.on_push is not None:
                    self.on_push(self.uin_points_layer.data)
                self.clear_data()

    def connect(self, on_push):
        self.on_push = on_push

    def set_pts(self, pts: npt.NDArray[np.int32]):
        for pt in pts:
            self.insert_pt(pt)

    def insert_pt(self, pt_tup: tuple) -> int:
        cnt_idx = self.lblinfo.lbl_im[pt_tup].item()
        self.pt_to_cnt_idx[pt_tup] = cnt_idx
        self.cnt_idx_to_pt[cnt_idx].append(pt_tup)
        self.cnt_idx_to_pt_history[cnt_idx].add(pt_tup)
        return cnt_idx

    def remove_pt(self, pt_tup: tuple) -> int:
        cnt_idx = self.pt_to_cnt_idx.pop(pt_tup)
        self.cnt_idx_to_pt[cnt_idx].remove(pt_tup)
        return cnt_idx

    def clear_data(self):
        self.last_pts: npt.NDArray[np.int32] = None
        self.pt_to_paintid = {}
        self.pt_to_cnt_idx = {}
        self.cnt_idx_to_pt = {i: [] for i in range(self.lblinfo.nlbl + 1)}
        self.cnt_idx_to_pt_history = {i: set() for i in range(self.lblinfo.nlbl + 1)}
        self.uin_points_layer.data = np.zeros((0, self.ndim), dtype=np.int32)
        self.uin_layer.data = np.zeros_like(self.uin_layer.data)
        self.uin_layer.refresh()

    def update_lblinfo(self, lblinfo: LblInfo):
        # TODO: allow regional update
        saved_pts = self.last_pts
        self.lblinfo = lblinfo
        self.clear_data()
        if saved_pts is not None:
            self.set_pts(saved_pts)  # this is to avoid using the previous threshold for new object

    def on_points_changed(self, event):
        pts = event.source.data
        if len(pts) > 0:
            pts = [self.im_layer.world_to_data(self.uin_points_layer.data_to_world(pt)) for pt in pts]
            pts = np.array(pts, dtype=np.float32)
            pts = pts.round().astype(np.int32)

            # filter off points that are outside and force points onto grid
            im_shape = self.im_arr.shape
            off_screen = np.zeros((pts.shape[0],), dtype=np.bool)
            for i in range(len(im_shape)):
                off_screen = off_screen | (pts[:, i] < 0) | (pts[:, i] >= im_shape[i])
            pts = pts[~off_screen]
        else:
            pts = np.zeros((0, 3), dtype=np.int32)

        self.update_uin_layer(pts)

    def update_uin_layer(self, pts: npt.NDArray[np.int32]):
        lblinfo = self.lblinfo
        pts = np.unique(pts, axis=0)

        if self.last_pts is not None:
            diff_pts1 = np_algorithms.setdiff2d(self.last_pts, pts)
            diff_pts2 = np_algorithms.setdiff2d(pts, self.last_pts)
        else:
            diff_pts1 = np.zeros((0, pts.shape[1]), dtype=pts.dtype)
            diff_pts2 = pts
        # all_changed_pts_map = {}
        self.last_pts = pts

        changed_cnts = set()
        for pt in diff_pts1:
            pt_tup = tuple(pt)
            changed_cnt = self.remove_pt(pt_tup)
            changed_cnts.add(changed_cnt)
            self.pt_to_paintid.pop(pt_tup)
        for pt in diff_pts2:
            pt_tup = tuple(pt)
            changed_cnt = self.insert_pt(pt_tup)
            changed_cnts.add(changed_cnt)
            self.pt_to_paintid[pt_tup] = lblinfo.assign_id()
        if 0 in changed_cnts:
            changed_cnts.remove(0)  # we don't need to split background
        if len(changed_cnts) == 0:
            return

        # now we have a list of changed contours, we want to modify the original image regions
        # where these changed contours overlap
        uin_im = self.uin_layer.data
        for cnt_idx in changed_cnts:
            print(cnt_idx, 'changed')
            contour_slices = lblinfo.cnt_slices[cnt_idx - 1]
            contour_pts: list = self.cnt_idx_to_pt[cnt_idx]
            all_contour_pts = np.array(contour_pts, dtype=np.int32)
            all_pts_history = list(self.cnt_idx_to_pt_history[cnt_idx])
            all_changed_pts = np.array(all_pts_history, dtype=np.int32)
            changed_pts_slices, _ = construct_slices(all_changed_pts, width=999)

            paintids = np.array([self.pt_to_paintid[pt_tup] for pt_tup in contour_pts], dtype=np.int32)

            for i in range(len(changed_pts_slices)):
                slices = take_intersect(contour_slices, changed_pts_slices[i])
                offset_mask = lblinfo.lbl_im[slices] == cnt_idx  # this is the contour the points split on

                if len(contour_pts) == 0:
                    splitted = np.zeros(offset_mask.shape, dtype=np.int32)
                else:
                    if are_in_slices(all_contour_pts, slices).sum() == 0:
                        splitted = np.zeros(offset_mask.shape, dtype=np.int32)
                    else:
                        stpt = np.array(tuple(s.start for s in slices), dtype=np.int32)
                        splitted = np_algorithms.voronoi_ndarray(offset_mask.shape, all_contour_pts - stpt[None, :])
                        splitted = splitted * offset_mask

                        splitted = paintids[splitted]
                uin_im[slices] = (uin_im[slices] * ~offset_mask) + (splitted * offset_mask)
        self.uin_layer.refresh()


def setup_annotation_callbacks(viewer: napari.Viewer,
                               record_manager: AnnotationRecordManager,
                               masks: list[dict[str, Any]]):
    im_arr: npt.NDArray = masks[0]['arr']

    uinb_mask = UserInputBuffer(viewer, record_manager, masks, name='uin_canvas')
    lblinfo_mask = LblInfo(im_arr > .44, 1)
    uinb_mask.update_lblinfo(lblinfo_mask)

    uinb = UserInputBuffer(viewer, record_manager, masks, name='uin_buffer', bind_key='ctrl+d')
    lblinfo = LblInfo(im_arr > .44, 1)
    uinb.update_lblinfo(lblinfo)

    def on_push(pts):
        layer = uinb_mask.uin_points_layer
        layer.data = np.concatenate((layer.data, pts), axis=0).astype(np.int32)
    uinb.connect(on_push)




import napari
import cvpl_tools.ome_zarr.io as cvpl_ome_zarr_io
import cvpl_tools.ome_zarr.napari.zarr_viewer as cvpl_napari_zarr
import numpy as np
import numpy.typing as npt
import dask.array as da
import json
from cvpl_tools.ome_zarr.napari.annotation_record_manager import AnnotationRecordManager
import cvpl_tools.ome_zarr.napari.annotation_callbacks as annotation_callbacks


def read_scale(ome_zarr_path: str, use_zip: bool = None) -> npt.NDArray[np.float32]:
    """Returns the scale of the first level of the image as a numpy array"""
    zarr_group = cvpl_ome_zarr_io.load_zarr_group_from_path(ome_zarr_path, mode='r', use_zip=use_zip)
    metadata = zarr_group.attrs['multiscales'][0]
    scale = metadata['datasets'][0]['coordinateTransformations'][0]['scale']
    return np.array(scale, dtype=np.float32)


def read_arr_shape(ome_zarr_path: str, use_zip: bool = None) -> npt.NDArray[np.int32]:
    zarr_group = cvpl_ome_zarr_io.load_zarr_group_from_path(ome_zarr_path, mode='r', use_zip=use_zip)['0']
    arr = da.from_zarr(zarr_group)
    return np.array(arr.shape, dtype=np.int32)


def mask_from_path(viewer: napari.Viewer, mask_path, viewer_args: dict, im_shape, dtype):
    if mask_path is None:
        # create new array
        mask_arr = np.zeros(im_shape, dtype=dtype)
    else:
        zarr_group = cvpl_ome_zarr_io.load_zarr_group_from_path(mask_path, level=0)
        mask_arr = da.from_zarr(zarr_group).compute()
    layer = viewer.add_labels(mask_arr, **viewer_args)
    return mask_arr, layer


def annotate(ome_zarr_path: str, output_dir: str):
    im_scale = read_scale(ome_zarr_path)
    im_scale[0] *= 2  # TODO: remove this in a real dataset
    im_shape = read_arr_shape(ome_zarr_path)
    im_shape[1:] //= 4
    im_shape_slices = tuple(slice(0, s.item()) for s in im_shape)
    record_manager = AnnotationRecordManager(output_dir)
    try:
        viewer = napari.Viewer(ndisplay=2)

        im_zarr_group = cvpl_ome_zarr_io.load_zarr_group_from_path(ome_zarr_path, level=0, mode='r')
        im_arr = np.clip(da.from_zarr(im_zarr_group)[im_shape_slices].compute() / 1000., 0., 1.)
        im_layer = viewer.add_image(im_arr, name='lightsheet', scale=im_scale, contrast_limits=(0., 1.))

        # load images
        sid = record_manager.get_sid()
        # mask_path = f'{output_dir}/annotation_mask_{sid - 1}'
        # mask, mask_layer = mask_from_path(viewer,
        #                                   mask_path if sid > 1 else None,
        #                                   dict(
        #                                       name='annotation_mask',
        #                                       scale=im_scale,
        #                                   ),
        #                                   im_shape,
        #                                   dtype=np.int32)
        completed_path = f'{output_dir}/completed_{sid - 1}'
        completed, completed_layer = mask_from_path(viewer,
                                                    completed_path if sid > 1 else None,
                                                    dict(
                                                        name='completed_mask',
                                                        scale=(im_scale * 64).astype(np.int64),
                                                        translate=tuple(s // 2 for s in im_shape)
                                                    ),
                                                    im_shape // 64,
                                                    dtype=np.uint8)
        print(im_shape // 64, im_scale * 64)

        # register images to record manager
        masks = [
            dict(
                arr=im_arr,
                name='im',
                path=None,
                is_numpy=True,
                on_update=lambda: None,
                on_close=lambda: None,  # TODO: change this to save onto disk
                layer=im_layer
            ),
            # dict(
            #     arr=mask,
            #     name='annotated_mask',
            #     path=mask_path,
            #     is_numpy=True,
            #     on_update=lambda: mask_layer.refresh(),
            #     on_close=lambda: None,  # TODO: change this to save onto disk
            #     layer=mask_layer
            # ),
            dict(
                arr=completed,
                name='completed_mask',
                path=completed_path,
                is_numpy=True,
                on_update=lambda: completed_layer.refresh(),
                on_close=lambda: None,  # TODO: change this to save onto disk
                layer=completed_layer
            )
        ]
        record_manager.register_masks(masks)

        # setup editing callbacks for mouse and keyboard
        annotation_callbacks.setup_annotation_callbacks(viewer, record_manager, masks)

        viewer.title = 'zarr_annotator'
        napari.run()
    finally:
        record_manager.close()



"""
This file defines how to read images whose representations are n-dimensional arrays, occasionally with metadata
associated with them.
"""

from abc import ABC, abstractmethod
from enum import Enum, auto

import numpy as np
import numpy.typing as npt
import dask.array as da


class ImageFileTypes(Enum):
    PNG = auto()
    TIF = auto()
    NIFTI = auto()
    ZARR = auto()
    OME_ZARR = auto()


class NumpyInterface(ABC):

    @abstractmethod
    def get_numpy(self) -> npt.NDArray:
        raise NotImplementedError()

    @abstractmethod
    def set_numpy(self, arr: npt.NDArray):
        raise NotImplementedError()

    @abstractmethod
    @property
    def dtype(self) -> np.dtype:
        raise NotImplementedError()

    @abstractmethod
    @property
    def ndim(self) -> int:
        raise NotImplementedError()

    @abstractmethod
    @property
    def shape(self) -> tuple[int, ...]:
        raise NotImplementedError()


class DaskInterface(ABC):
    pass
